/*	$Csoft: cgi.c,v 1.16 2005/07/06 07:36:37 vedge Exp $	*/

/*
 * Copyright (c) 2003, 2004 CubeSoft Communications, Inc.
 * <http://www.csoft.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <libcgi/cgi.h>
#include <libcgi/mime.h>

#include <config/localedir.h>
#include <config/sysconfdir.h>
#include <config/have_pctr.h>

#include <sys/stat.h>
#include <sys/mman.h>

#include <ctype.h>
#include <signal.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <locale.h>
#include <fcntl.h>

#ifdef HAVE_PCTR
#include <machine/pctr.h>
#endif

struct cgi *cgi;
unsigned int cgi_nqueries = 0;
static char cgi_errmsg[256];
static volatile sig_atomic_t cgi_got_sigpipe = 0;
static volatile sig_atomic_t cgi_got_sigterm = 0;
#ifdef HAVE_PCTR
static pctrval cgi_pctrv;
#endif

const char *cgi_loglvl_names[] = {
	"emerg",
	"alert",
	"crit",
	"err",
	"warn",
	"notice",
	"info",
	"debug"
};

/* Escape characters as described in RFC1738. */
char *
cgi_url_escape(struct cgi_query *q, const char *url)
{
	char hex[3];
	size_t len;
	const u_char *sp;
	char *dst, *dp;

	len = strlen(url)+1;
	dp = dst = Malloc(len);
	hex[2] = '\0';
	for (sp = url; *sp != '\0'; dp++, sp++) {
		const char *url_unsafe = "<>\"#%{}|\\^~[]`", *p;
		int unsafe = 0;

		if (!isgraph(*sp)) {
			unsafe++;
		} else {
			for (p = &url_unsafe[0]; *p != '\0'; p++) {
				if (*p == *sp) {
					unsafe++;
					break;
				}
			}
		}
		if (unsafe) {
			snprintf(hex, sizeof(hex), "%02x", *sp);
			len += 3;
			dst = Realloc(dst, len);
			dp[0] = '%';
			dp[1] = hex[0];
			dp[2] = hex[1];
			dp += 2;
		} else {
			*dp = *sp;
		}
	}
	*dp = '\0';
	return (dst);
}

/*
 * Unescape characters as described in RFC1738, except for NUL sequences
 * which are replaced by underscores.
 */
char *
cgi_url_unescape(struct cgi_query *q, const char *url)
{
	char hex[3];
	const char *sp;
	char *dst, *dp;
	unsigned int n;

	dp = dst = Malloc(strlen(url)+2);
	hex[2] = '\0';
	for (sp = url; *sp != '\0'; dp++, sp++) {
		if (sp[0] == '%' && isxdigit(sp[1]) && isxdigit(sp[2])) {
			hex[0] = sp[1];
			hex[1] = sp[2];
			n = (unsigned int)strtoul(hex, NULL, 16);
			if (n == '\0') {
				*dp = '_';
				cgi_log(LOG_ERR, "Replaced NUL escape in URL");
			} else {
				*dp = n;
			}
			sp += 2;
		} else if (sp[0] == '+') {
			*dp = ' ';
		} else {
			*dp = sp[0];
		}
	}
	*dp = '\0';
	return (dst);
}

/* Parse application/x-www-form-urlencoded arguments. */
static void
cgi_url_decode(struct cgi_query *q, char *qstring, enum cgi_argument_type t)
{
	struct cgi_argument *arg;
	char *s, *qs = qstring;
	char *name, *value;
	int nargs;

	for (nargs = 0;
	    (s = strsep(&qs, "&")) != NULL && nargs < CGI_MAX_ARGS;
	     nargs++) {
		arg = Malloc(sizeof(struct cgi_argument));
		if ((name = strsep(&s, "=")) != NULL &&
		    name[0] != '\0') {
			if (strlcpy(arg->key, name, sizeof(arg->key)) >=
			    sizeof(arg->key)) {
				cgi_log(LOG_ERR, "Key is too big: `%s'",
				    arg->key);
				free(arg);
				continue;
			}
			if ((value = strsep(&s, "=")) != NULL) {
				arg->value = cgi_url_unescape(q, value);
			} else {
				arg->value = Strdup("");
			}
		    	arg->type = t;
			TAILQ_INSERT_TAIL(&q->args, arg, args);
		}
	}
}

/* Parse POSTDATA if its Content-type is known. */
static void
cgi_decode_post(struct cgi_query *q)
{
	char *cntype, *s, *buf;
	size_t cnlength;

	if ((cntype = getenv("CONTENT_TYPE")) == NULL)
		return;
	if ((s = getenv("CONTENT_LENGTH")) == NULL ||
	    (cnlength = (size_t)strtol(s, NULL, 10)) == 0)
		return;

	if (strcmp(cntype, "application/x-www-form-urlencoded") == 0) {
		buf = Malloc(cnlength);
		if (fread(buf, cnlength, 1, stdin) < 1) {
			cgi_log(LOG_ERR, "Error reading POSTDATA from stdin.");
			free(buf);
			return;
		}
		cgi_url_decode(q, buf, CGI_POST_ARGUMENT);
		free(buf);
	} else {
		cgi_log(LOG_ERR, "Unknown POST Content-type: `%s'", cntype);
	}
}

static void
cgi_decode_cookies(struct cgi_query *q, char *s)
{
	struct cgi_cookie *ck;
	char *sp = s, *sc;

	while ((sc = strsep(&sp, ";")) != NULL) {
		char *skey = strsep(&sc, "=");
		char *sval = strsep(&sc, "=");

		if (skey == NULL || skey[0] == '\0' || sval == NULL)
			continue;

		ck = Malloc(sizeof(struct cgi_cookie));
		strlcpy(ck->name, skey, sizeof(ck->name));
		strlcpy(ck->value, sval, sizeof(ck->value));
		ck->expire[0] = '\0';
		ck->domain[0] = '\0';
		ck->path[0] = '\0';
		ck->secure = 0;
		TAILQ_INSERT_HEAD(&q->cookies, ck, cookies);
	}
}

char *
cgi_get_cookie(struct cgi_query *q, const char *name)
{
	struct cgi_cookie *ck;

	TAILQ_FOREACH(ck, &q->cookies, cookies) {
		if (strcmp(ck->name, name) == 0)
			return (ck->value);
	}
	return (NULL);
}

struct cgi_cookie *
cgi_set_cookie(struct cgi_query *q, const char *name, const char *fmt, ...)
{
	struct cgi_cookie *ck;
	va_list ap;

	ck = Malloc(sizeof(struct cgi_cookie));
	strlcpy(ck->name, name, sizeof(ck->name));
	ck->expire[0] = '\0';
	ck->domain[0] = '\0';
	ck->path[0] = '\0';
	ck->secure = 0;

	va_start(ap, fmt);
	vsnprintf(ck->value, sizeof(ck->value), fmt, ap);
	va_end(ap);

	TAILQ_INSERT_HEAD(&q->cookies, ck, cookies);
	return (ck);
}

/*
 * Accept a request from the web server; allocate a CGI query, parse
 * the script arguments, parse the cookies and negotiate the
 * language/character set.
 */
struct cgi_query *
cgi_accept(void)
{
	struct cgi_query *q;
	char **ap;
	char *oklang, *okcset;
	char **lang, **cset;
	char **avcset, **avlang;
	const char *rlang;
	char *s, *sep;

#ifdef HAVE_FASTCGI
	if (FCGI_Accept() < 0)
		return (NULL);
#else
	if (++cgi_nqueries > 1) {
		cgi_nqueries = 0;
		return (NULL);
	}
#endif
	q = Malloc(sizeof(struct cgi_query));
	strlcpy(q->lang, cgi->avail_langs[0], sizeof(q->lang));
	strlcpy(q->cset, cgi->avail_csets[0], sizeof(q->cset));
	q->res_type = NULL;
	TAILQ_INIT(&q->args);
	TAILQ_INIT(&q->cookies);

	/* Parse the list of acceptable languages. */
	oklang = getenv("HTTP_ACCEPT_LANGUAGE");
	for (ap = &q->accept_langs[0];
	     ap < &q->accept_langs[CGI_LANGS_MAX] &&
	     (*ap = strsep(&oklang, ","));) {
		if ((sep = strchr(*ap, ';')) != NULL)
			*sep = '\0';
		if (**ap != '\0')
			ap++;
	}
	*ap = NULL;

	/* Parse the list of acceptable character sets. */
	okcset = getenv("HTTP_ACCEPT_CHARSET");
	for (ap = &q->accept_csets[0];
	     ap < &q->accept_csets[CGI_CSETS_MAX] &&
	     (*ap = strsep(&okcset, ","));) {
		if ((sep = strchr(*ap, ';')) != NULL)
			*sep = '\0';
		if (**ap != '\0')
			ap++;
	}
	*ap = NULL;

	/* Decode the GET/POST information. */
	if ((s = getenv("QUERY_STRING")) != NULL && s[0] != '\0') {
		cgi_url_decode(q, s, CGI_GET_ARGUMENT);
	} else {
		cgi_decode_post(q);
	}
	q->script_name = getenv("SCRIPT_NAME");
	var_set("prog", q->script_name);
	
	/* Parse the cookies. */
	if ((s = getenv("HTTP_COOKIE")) != NULL && s[0] != '\0')
		cgi_decode_cookies(q, s);

	/* Negotiate the character set. */
	for (avcset = &cgi->avail_csets[0]; *avcset != NULL; avcset++) {
		for (cset = &q->accept_csets[0]; *cset != NULL; cset++) {
			if (strcasecmp(*cset, *avcset) == 0)
				break;
		}
		if (*avcset != NULL) {
			strlcpy(q->cset, *avcset, sizeof(q->cset));
			break;
		}
	}
	
	/* Negotiate and configure the language. */
	if ((rlang = cgi_get_arg(q, "lang")) != NULL &&
	    strlen(rlang) <= 3) {
		for (avlang = &cgi->avail_langs[0]; *avlang != NULL; avlang++) {
			if (strcmp(rlang, *avlang) == 0)
				break;
		}
		if (*avlang != NULL)
			strlcpy(q->lang, rlang, sizeof(q->lang));
	} else {
		for (lang = &q->accept_langs[0]; *lang != NULL; lang++) {
			for (avlang = &cgi->avail_langs[0]; *avlang != NULL;
			     avlang++) {
				if (strcasecmp(*lang, *avlang) == 0)
					break;
			}
			if (*avlang != NULL) {
				strlcpy(q->lang, *avlang, sizeof(q->lang));
				break;
			}
		}
	}
	setenv("LANGUAGE", q->lang, 1);
	var_set("lang", q->lang);
	return (q);
}

/* Negotiate the language and character set; write the script headers. */
void
cgi_begin(struct cgi_query *q, const char *res_type)
{
	struct mime_entity ment;
	struct cgi_cookie *ck;

	mime_init_entity(&ment);
	mime_add_header(&ment, "Content-type:%s;charset=%s", res_type, q->cset);
	TAILQ_FOREACH(ck, &q->cookies, cookies) {
		char ch[256];

		strlcpy(ch, "Set-Cookie: ", sizeof(ch));
		strlcat(ch, ck->name, sizeof(ch));
		strlcat(ch, "=", sizeof(ch));
		strlcat(ch, ck->value, sizeof(ch));
		if (ck->expire[0] != '\0') {
			strlcat(ch, "; expires=", sizeof(ch));
			strlcat(ch, ck->expire, sizeof(ch));
		}
		if (ck->domain[0] != '\0') {
			strlcat(ch, "; domain=", sizeof(ch));
			strlcat(ch, ck->domain, sizeof(ch));
		}
		if (ck->path[0] != '\0') {
			strlcat(ch, "; path=", sizeof(ch));
			strlcat(ch, ck->path, sizeof(ch));
		}
		if (ck->secure) {
			strlcat(ch, "; secure", sizeof(ch));
		}
		mime_add_header_str(&ment, ch);
	}
	mime_write(q, &ment);
	mime_destroy_entity(&ment);
	
	q->res_type = res_type;
	
#ifdef HAVE_PCTR
	cgi_pctrv = rdtsc();
#endif
}

/*
 * Return a pointer to the value of a cgi argument or NULL if it does not
 * exist. No length checking is done.
 */
const char *
cgi_get_arg(struct cgi_query *q, const char *key)
{
	struct cgi_argument *arg;

	TAILQ_FOREACH(arg, &q->args, args) {
		if (strcmp(arg->key, key) == 0)
			break;
	}
	if (arg == NULL) {
		error_set(_("The `%s' argument is missing."), key);
		return (NULL);
	}
	return (arg->value);
}

/*
 * Return a pointer to the value of a cgi argument, or NULL if it does not
 * exist or exceeds len-1 bytes.
 */
const char *
cgi_lget_arg(struct cgi_query *q, const char *key, size_t len)
{
	struct cgi_argument *arg;

	TAILQ_FOREACH(arg, &q->args, args) {
		if (strcmp(arg->key, key) == 0)
			break;
	}
	if (arg == NULL) {
		error_set(_("The `%s' argument is missing."), key);
		return (NULL);
	}
	if (strlen(arg->value) > len) {
		cgi_log(LOG_ERR, "`%s' argument is >%lu bytes.", key,
		    (unsigned long)len);
		error_set(_("The `%s' argument exceeds %lu bytes."), key,
		    (unsigned long)len);
		return (NULL);
	}
	return (arg->value);
}

/* Release the resources allocated by a CGI query and check for signals. */
void
cgi_end(struct cgi_query *q)
{
	struct cgi_argument *arg, *narg;
	struct cgi_cookie *c, *nc;
	struct var *var, *nvar;

	for (arg = TAILQ_FIRST(&q->args);
	     arg != TAILQ_END(&q->args);
	     arg = narg) {
		narg = TAILQ_NEXT(arg, args);
		free(arg->value);
		free(arg);
	}
	for (c = TAILQ_FIRST(&q->cookies);
	     c != TAILQ_END(&q->cookies);
	     c = nc) {
		nc = TAILQ_NEXT(c, cookies);
		free(c);
	}
	for (var = SLIST_FIRST(&cgi->vars);
	     var != SLIST_END(&cgi->vars);
	     var = nvar) {
		nvar = SLIST_NEXT(var, vars);
		if (var->global) {
			continue;
		}
		SLIST_REMOVE(&cgi->vars, var, var, vars);
		Free(var->value);
		free(var);
	}

	free(q);

	if (cgi_got_sigterm) {
		cgi_log(LOG_INFO, "Received SIGTERM; exiting");
		cgi_destroy(EX_OK, NULL);
	}
	if (cgi_got_sigpipe) {
		cgi_got_sigpipe = 0;
		cgi_log(LOG_INFO, "Client abort");
	}

#ifdef HAVE_FASTCGI
	FCGI_Finish();
#endif
}

static void
cgi_sigpipe(int sigraised)
{
	cgi_got_sigpipe++;
}

static void
cgi_sigterm(int sigraised)
{
	cgi_got_sigterm++;
#if 1
	cgi_destroy(EX_OK, NULL);
#endif
}

/* Initialize the CGI library. */
void
cgi_init(struct cgi *pcgi)
{
	extern struct cgi_filter varsubst_filter;
	extern char *__progname;

	setlocale(LC_ALL, "");
#ifdef ENABLE_NLS
	bindtextdomain("csoft-cgi", LOCALEDIR);
	textdomain("csoft-cgi");
#endif
	cgi = pcgi;

#ifndef HAVE_FASTCGI
	if (cgi->flags & CGI_PERSISTENT) {
		cgi_log(LOG_EMERG, "%s requires FastCGI. Recompile csoft-cgi "
		                   "with the --with-fastcgi option.\n",
				   __progname);
		exit(1);
	}
#else
	cgi_log(LOG_INFO, "%s is now accepting requests.", __progname);
#endif

	cgi->avail_csets[0] = "utf-8";			     /* Prefer UTF-8 */
	cgi->avail_csets[1] = "iso-8859-1";
	cgi->avail_csets[2] = NULL;
	TAILQ_INIT(&cgi->filters);
	SLIST_INIT(&cgi->vars);

	cgi_add_filter(&varsubst_filter);

	var_set_global("progname", __progname);

	signal(SIGTERM, cgi_sigterm);
	signal(SIGUSR1, cgi_sigterm);
	signal(SIGPIPE, cgi_sigpipe);
}

/* Register a content filter. */
void
cgi_add_filter(struct cgi_filter *fil)
{
	TAILQ_INSERT_TAIL(&cgi->filters, fil, filters);
}

/* Clean up and exit. */
void
cgi_destroy(int excode, const char *fmt, ...)
{
	if (cgi->destroy_fn != NULL)
		cgi->destroy_fn();

	if (fmt != NULL) {
		char msg[BUFSIZ];
		va_list ap;

		va_start(ap, fmt);
		vsnprintf(msg, sizeof(msg), fmt, ap);
		va_end(ap);

		cgi_log(LOG_ERR, "%s; exiting\n", msg);

		if (cgi->flags & CGI_HTML_ERRORS) {
			printf(_("<h1>Script failure:</h1>"));
			printf("%s<br>", msg);
			printf(_("Please report this error to "
			         "&lt;<a href='mailto:%s'>%s</a>&gt;"
				 "<br>"), msg, getenv("SERVER_ADMIN"));
			exit(EX_OK);
		}
	}
	exit(excode);
}

void *
cgi_malloc(size_t len)
{
	void *buf;

	buf = malloc(len);
	if (buf == NULL) {
		cgi_destroy(EX_TEMPFAIL, _("Out of memory"));
	}
	return (buf);
}

void *
cgi_realloc(void *p, size_t len)
{
	void *buf;

	buf = realloc(p, len);
	if (buf == NULL) {
		cgi_destroy(EX_TEMPFAIL, _("Out of memory"));
	}
	return (buf);
}

void
cgi_free(void *p)
{
	if (p != NULL)
		free(p);
}

char *
cgi_strdup(const char *s)
{
	char *sd;

	if ((sd = strdup(s)) == NULL) {
		cgi_destroy(EX_TEMPFAIL, _("Out of memory"));
	}
	return (sd);
}

/*
 * Write data to the standard output, applying output filters registered for
 * the query's content-type, if it is set.
 */
ssize_t
cgi_write(struct cgi_query *q, void *buf, size_t size)
{
	struct cgi_filter *fil;

	if (q->res_type != NULL) {
		TAILQ_FOREACH(fil, &cgi->filters, filters) {
			if (fil->type == CGI_OUTPUT_FILTER &&
			    strcmp(fil->content_type, q->res_type) == 0) {
				size = fil->func(q, &buf, size);
			}
		}
		if (fil == NULL)
			goto plain;
	} else {
plain:
		if (fwrite(buf, size, 1, stdout) < 1) {
			cgi_log(LOG_ERR, feof(stdout) ?
			    _("End of file on stdout") :
			    _("Write error on stdout"));
			abort();
			return (-1);
		}
	}
	return ((ssize_t)size);
}

/*
 * Emit an error message to the error output. Format the message into a
 * fixed-size buffer in order to limit the size of log entries.
 */
void
cgi_log(enum cgi_loglvl level, const char *fmt, ...)
{
	char errbuf[256];
	va_list ap;

	va_start(ap, fmt);
	vsnprintf(errbuf, sizeof(errbuf), fmt, ap);
	va_end(ap);
	strlcat(errbuf, "\n", sizeof(errbuf));

#ifdef HAVE_FASTCGI
	{
		char file[FILENAME_MAX];
		extern char *__progname;
		int fd;

		strlcpy(file, __progname, sizeof(file));
		strlcat(file, ".log", sizeof(file));

		fd = open(file, O_WRONLY|O_APPEND|O_CREAT, 0600);
		write(fd, errbuf, strlen(errbuf));
		close(fd);
	}
#else
	fprintf(stderr, "[%s] %s\n", cgi_loglvl_names[level], errbuf);
#endif
}

/*
 * Write a formatted string to the standard output, assuming the formatted
 * text is consistent with the current output Content-Type.
 */
void
cgi_printf(struct cgi_query *q, const char *fmt, ...)
{
	char *buf;
	va_list ap;

	va_start(ap, fmt);
	if (vasprintf(&buf, fmt, ap) == -1) {
		cgi_destroy(EX_TEMPFAIL, _("Out of memory"));
	}
	va_end(ap);

	cgi_write(q, buf, strlen(buf));
	free(buf);
}

void
error_set(const char *fmt, ...)
{
	va_list args;
	
	va_start(args, fmt);
	vsnprintf(cgi_errmsg, sizeof(cgi_errmsg), fmt, args);
	va_end(args);
}

char *
error_get(void)
{
	return (cgi_errmsg);
}

/*
 * Search for the named data file and copy its absolute path name into a
 * fixed-size buffer.
 */
int
cgi_find_datafile(const char *name, char *dst, size_t dst_len)
{
	extern char *__progname;
	char path[MAXPATHLEN];
	struct stat sta;
	
	strlcpy(path, __progname, sizeof(path));
	strlcat(path, ".conf/", sizeof(path));
	strlcat(path, name, sizeof(path));
	if (stat(path, &sta) == 0 &&
	    strlcpy(dst, path, dst_len) < dst_len)
		return (0);
	
	strlcpy(path, SYSCONFDIR, sizeof(path));
	strlcat(path, "/", sizeof(path));
	strlcat(path, __progname, sizeof(path));
	strlcat(path, "/", sizeof(path));
	strlcat(path, name, sizeof(path));
	if (stat(path, &sta) == 0 &&
	    strlcpy(dst, path, dst_len) < dst_len)
		return (0);

	error_set(_("The data file `%s' was not found."), name);
	return (-1);
}

/*
 * Map the named control file into memory for further parsing.
 * If the file is empty, set *p to NULL and return successfully.
 */
int
cgi_map_datafile(const char *name, char **p, size_t *len)
{
	char path[MAXPATHLEN];
	char *map;
	int fd;
	
	if (cgi_find_datafile(name, path, sizeof(path)) == -1) {
		return (-1);
	}
	if ((fd = open(path, O_RDONLY)) == -1) {
		error_set("%s: %s", path, strerror(errno));
		return (-1);
	}
	if ((*len = (size_t)lseek(fd, 0, SEEK_END)) == 0) {
		*p = NULL;
		goto out;
	}
	if ((map = mmap(NULL, *len, PROT_READ|PROT_WRITE, MAP_FILE|MAP_PRIVATE,
	    fd, 0)) == MAP_FAILED) {
		error_set("mmap %s: %s", path, strerror(errno));
		goto fail;
	}
	*p = map;
out:
	close(fd);
	return (0);
fail:
	close(fd);
	return (-1);
}

/*
 * Return those variables with a key that starts with the given pattern
 * (used for things like checkboxes).
 */
struct cgi_keyset *
cgi_get_keyset(struct cgi_query *q, const char *keypat)
{
	size_t patlen = strlen(keypat);
	struct cgi_argument *arg;
	struct cgi_keyset *set;

	set = Malloc(sizeof(struct cgi_keyset));
	set->nents = 0;
	set->ents = Malloc(1);

	TAILQ_FOREACH(arg, &q->args, args) {
		if (strncmp(arg->key, keypat, patlen) != 0)
			continue;

		set->ents = Realloc(set->ents, (set->nents+1) * sizeof(char *));
		set->ents[set->nents++] = cgi_url_unescape(q, arg->key+patlen);
	}
	return (set);
}

void
cgi_free_keyset(struct cgi_keyset *set)
{
	int i;

	for (i = 0; i < set->nents; i++) {
		free(set->ents[i]);
	}
	free(set->ents);
	free(set);
}

void
cgi_get_time(void)
{
#ifdef HAVE_PCTR
	var_set("cgi_time_count", _("Request served in %.02fM cycles."),
	    (float)(rdtsc() - cgi_pctrv) / 1000000);
#endif
}

