/*	$Csoft: table.c,v 1.4 2004/01/03 04:10:19 vedge Exp $	*/

/*
 * Copyright (c) 2003, 2004 CubeSoft Communications, Inc.
 * <http://www.csoft.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <libcgi/cgi.h>
#include <libcgi/table.h>

#include <sys/mman.h>

#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

/* Initialize a table. */
void
table_init(struct table *ta)
{
	int i;

	ta->data = NULL;
	ta->data_len = 0;
	for (i = 0; i < TABLE_NBUCKETS; i++) {
		SLIST_INIT(&ta->buckets[i].ents);
	}
	SLIST_INIT(&ta->ents);
}

/* Release the resources allocated by a table. */
void
table_destroy(struct table *ta)
{
	struct table_ent *ent, *nent;
	int i;

	for (i = 0; i < TABLE_NBUCKETS; i++) {
		for (ent = SLIST_FIRST(&ta->buckets[i].ents);
		     ent != SLIST_END(&ta->buckets[i].ents);
		     ent = nent) {
			nent = SLIST_NEXT(ent, bents);
			free(ent);
		}
	}
	munmap(ta->data, ta->data_len);
}

static __inline u_int
table_hash(const char *key)
{
	u_int h;
	u_char *p;

	for (h = 0, p = (u_char *)key; *p != '\0'; p++) {
		h = 31 * h + *p;
	}
	return (h % TABLE_NBUCKETS);
}

/* Look up a named table entry. */
struct table_ent *
table_lookup(struct table *ta, const char *key, u_int h)
{
	struct table_ent *ent;

	SLIST_FOREACH(ent, &ta->buckets[h].ents, bents) {
		if (strcmp(key, ent->key) == 0)
			break;
	}
	if (ent == NULL) {
		error_set(_("Element `%s' was not found in the table."), key);
	}
	return (ent);
}

/* Return a pointer to the given table entry field. */
char *
table_lookup_field(struct table *ta, const char *key, int field)
{
	struct table_ent *ent;

	if ((ent = table_lookup(ta, key, table_hash(key))) != NULL) {
		if (field >= ta->nfields) {
			error_set(_("Element `%s' has no %d field."), key,
			    field);
			return (NULL);
		}
		return (ent->fields[field]);
	}
	error_set(_("Element `%s' was not found in the table."), key);
	return (NULL);
}

/* Copy a named table entry field to a fixed-size buffer. */
int
table_copy_field(struct table *ta, const char *key, int field, char *buf,
    size_t buf_len)
{
	struct table_ent *ent;

	if ((ent = table_lookup(ta, key, table_hash(key))) != NULL) {
		if (field >= ta->nfields) {
			error_set(_("Element `%s' has no %d field."), key,
			    field);
			return (-1);
		}
		if (strlcpy(buf, ent->fields[field], buf_len) >= buf_len) {
			error_set(_("The string is too large for the buffer."));
			return (-1);
		}
		return (0);
	}
	return (-1);
}

/* Insert a new table entry. */
int
table_insert(struct table *ta, struct table_ent *ent)
{
	u_int h = table_hash(ent->key);

	if (table_lookup(ta, ent->key, h) != NULL) {
		error_set(_("Element `%s' exists in table."), ent->key);
		return (-1);
	}
	SLIST_INSERT_HEAD(&ta->buckets[h].ents, ent, bents);
	SLIST_INSERT_HEAD(&ta->ents, ent, ents);
	ta->buckets[h].nents++;
	return (0);
}

/* Remove a named table entry. */
int
table_remove(struct table *ta, const char *key)
{
	u_int h = table_hash(key);
	struct table_ent *ent;

	if ((ent = table_lookup(ta, key, h)) == NULL) {
		error_set(_("Element `%s' not found in table."), key);
		return (-1);
	}
	SLIST_REMOVE(&ta->buckets[h].ents, ent, table_ent, bents);
	SLIST_REMOVE(&ta->ents, ent, table_ent, ents);
	free(ent);
	return (0);
}

/*
 * Separate the fields of a table line. Fail if there are not enough fields,
 * but ignore extra fields if there are any. The field pointers of entries
 * point into the table's mmaped area, to avoid unnecessary copies.
 */
static int
table_parse_line(struct table *ta, char *s)
{
	struct table_ent *nent;
	char *key, **p;

	if ((key = strsep(&s, ":")) == NULL) {
		error_set(_("Missing key"));
		return (-1);
	}
	if ((nent = malloc(sizeof(struct table_ent))) == NULL) {
		error_set(_("Out of memory"));
		return (-1);
	}
	nent->key = key;
	for (p = &nent->fields[0];
	     p < &nent->fields[ta->nfields] &&
	     (*p = strsep(&s, ":")) != NULL; ) {
		if (**p != '\0')
			p++;
	}
	if (p < &nent->fields[ta->nfields-1]) {
		error_set(_("Missing fields (<%d)"), ta->nfields);
		goto fail_free;
	}
	if (table_insert(ta, nent) == -1) {
		goto fail_free;
	}
	return (0);
fail_free:
	free(nent);
	return (-1);
}

/*
 * Load a table from the given data file. Leading spacing characters and
 * characters between '#' and end of line are ignored.
 */
struct table *
table_load(const char *name, int nfields)
{
	struct table *ta;
	char *data, *p, *eol;
	size_t len;
	u_int nline = 0;

	if (cgi_map_datafile(name, &data, &len) == -1)
		return (NULL);

	if ((ta = malloc(sizeof(struct table))) == NULL) {
		error_set(_("Out of memory"));
		return (NULL);
	}
	table_init(ta);
	ta->data = data;
	ta->data_len = len;
	ta->nfields = nfields;

	if (data != NULL) {
		for (p = &data[0]; *p != '\0'; p++) {
			while (isspace(*p)) {
				p++;
			}
			if (*p == '#') {
				while (*(++p) != '\n')
				    ;;
			}
			for (eol = p;
			     *eol != '\n' && *eol != '\0';
			     eol++) {
				if (*eol == '#') {
					for (eol--; isspace(*eol); eol--)
					    ;;
					eol++;
					break;
				}
			}
			*eol = '\0';
			if (p < eol) {
				char *sp;

				for (sp = &eol[-1]; isspace(*sp); sp--)
				    ;;
				sp[1] = '\0';
				if (table_parse_line(ta, p) == -1)
					cgi_log(LOG_WARNING, "%s:%u: %s",
					    name, nline, error_get());
			}
			p = eol;
			nline++;
		}
	}
	return (ta);
}

/* Save the contents of the given table to the named file. */
int
table_save(struct table *ta, const char *path)
{
	struct table_ent *ent;
	FILE *f;
	int i;

	if ((f = fopen(path, "w")) == NULL) {
		error_set("%s: %s", path, error_get());
		return (-1);
	}
	TABLE_FOREACH(ent, ta) {
		fputs(ent->key, f);
		for (i = 0; i < ta->nfields; i++) {
			fputc(':', f);
			fputs(ent->fields[i], f);
		}
		fputc('\n', f);
	}
	fclose(f);
	return (0);
}

