/*	$Csoft: textbuf.c,v 1.3 2004/01/03 04:10:19 vedge Exp $	*/

/*
 * Copyright (c) 2003, 2004 CubeSoft Communications, Inc.
 * <http://www.csoft.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <libcgi/cgi.h>

#include <ctype.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>

#include "textbuf.h"

#define TEXT_BUFFER_GROW 2048

static size_t	utf8_to_latin1(struct textbuf *, const char *);

static const struct {
	char	 *name;
	size_t	(*conv_fn)(struct textbuf *, const char *);
} convs[] = {
	{ "ISO-8859-1",	utf8_to_latin1 }
};
static const int nconvs = sizeof(convs) / sizeof(convs[0]);

/* Initialize a new text buffer. */
void
textbuf_init(struct textbuf *te, size_t len, const char *encoding)
{
	strlcpy(te->encoding, encoding, sizeof(te->encoding));
	te->buf = Malloc(len);
	te->buf_len = len;
	te->len = 0;
}

/* Release a text buffer. */
void
textbuf_destroy(struct textbuf *te)
{
	free(te->buf);
}

/* Append a NUL-terminated string to a text buffer without converting. */
size_t
textbuf_append(struct textbuf *te, const char *s)
{
	size_t len;

	len = strlen(s);
	te->len += len;
	if (te->len > te->buf_len) {
		te->buf_len = te->len + TEXT_BUFFER_GROW;
		te->buf = Realloc(te->buf, te->buf_len);
	}
	memcpy(&te->buf[te->len-len], s, len);
	return (len);
}

/* Append a formatted string to a text buffer. */
size_t
textbuf_append_fmt(struct textbuf *te, const char *fmt, ...)
{
	char buf[1024];
	va_list ap;

	va_start(ap, fmt);
	vsnprintf(buf, sizeof(buf), fmt, ap);
	va_end(ap);
	return (textbuf_append(te, buf));
}

/* Calculate the length of a UTF-8 sequence in bytes. */
static __inline__ int
utf8_length(u_char ch)
{
	if ((ch >> 7) == 0) {
		return (1);
	} else if (((ch & 0xe0) >> 5) == 0x6) {
		return (2);
	} else if (((ch & 0xf0) >> 4) == 0xe) {
		return (3);
	} else if (((ch & 0xf8) >> 3) == 0x1e) {
		return (4);
	} else if (((ch & 0xfc) >> 2) == 0x3e) {
		return (5);
	} else if (((ch & 0xfe) >> 1) == 0x7e) {
		return (6);
	} else {
		return (-1);
	}
}

/* Try to convert Latin-1 UTF-8 characters to their ISO-8859-1 equivalent. */
static size_t
utf8_to_latin1(struct textbuf *te, const char *s)
{
	const u_char *sp;
	size_t len = strlen(s);
	u_int32_t c;

	if (te->len+len > te->buf_len) {
		te->buf_len = len + TEXT_BUFFER_GROW;
		te->buf = Realloc(te->buf, te->buf_len);
	}
	for (sp = &s[0]; *sp != '\0'; ) {
		int ntrail = 0;

		if (*sp < 0x80) {
			c = *(sp++);
		} else if (*sp < 0xc0) {
			c = '!';
			sp++;
		} else if (*sp < 0xe0) {
			c = *(sp++) & 0x1f;
			ntrail = 1;
		} else if (*sp < 0xf0) {
			c = *(sp++) & 0x0f;
			ntrail = 2;
		} else if (*sp < 0xf8) {
			c = *(sp++) & 0x07;
			ntrail = 3;
		} else {
			c = '?';
			sp++;
		}
		for (; ntrail > 0; ntrail--) {
			if ((*sp & 0xc0) != 0x80) {
				c = 'X';
				sp++;
				break;
			}
			c <<= 6;
			c |= *(sp++) & 0x3f;
		}
		if (c <= 0xff) {
			te->buf[te->len++] = c;
		} else {
			te->buf[te->len++] = '?';
		}
	}
	return (len);
}

/* Append an UTF-8 string to a text buffer, converting as necessary. */
size_t
textbuf_append_utf8(struct textbuf *te, const char *s)
{
	int i;
	
	for (i = 0; i < nconvs; i++) {
		if (strcasecmp(convs[i].name, te->encoding) == 0) {
			return (convs[i].conv_fn(te, s));
		}
	}
	return (textbuf_append(te, s));
}

/* Append a single character to a text buffer without converting. */
size_t
textbuf_append_char(struct textbuf *te, char c)
{
	if (++te->len > te->buf_len) {
		te->buf_len = te->len + TEXT_BUFFER_GROW;
		te->buf = Realloc(te->buf, te->buf_len);
	}
	te->buf[te->len-1] = c;
	return (1);
}

