/*	$Csoft: mail.c,v 1.9 2004/01/21 16:52:25 vedge Exp $	*/

/*
 * Copyright (c) 2003, 2004 CubeSoft Communications, Inc.
 * <http://www.csoft.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <libcgi/cgi.h>
#include <libcgi/mime.h>
#include <libcgi/html.h>
#include <libcgi/table.h>
#include <libcgi/textbuf.h>

#include <sys/stat.h>
#include <sys/wait.h>

#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <regex.h>
#include <stdlib.h>
#include <time.h>
#include <unistd.h>

#include "pathnames.h"

static void mail_destroy(void);

struct cgi mail_cgi = {
	N_("Mailer script"),
	N_("Copyright (c) 2003 CubeSoft Communications, Inc."),
	{ "en", "fr", "no", NULL },
	CGI_HTML_ERRORS,
	mail_destroy
};
struct table *Trcpts;

#define SENDER_MAX	64
#define RCPT_MAX	32
#define SUBJECT_MAX	128
#define MESSAGE_MAX	4096
#define XVAL_MAX	128
#define FVAL_MAX	256

/* Verify the format of the sender e-mail address. */
static int
valid_sender(const char *email)
{
	regmatch_t rm[2];
	regex_t reg;

	if (regcomp(&reg,
	    "^[a-zA-Z0-9_.-]{1,64}@[a-zA-Z0-9.-]{2,60}"
	    ".[a-zA-Z]{2,4}$", REG_EXTENDED) != 0) {
		cgi_log(LOG_ERR, "regcomp failed");
		return (0);
	}
	if (regexec(&reg, email, 2, rm, 0) != 0) {
		return (0);
	}
	regfree(&reg);
	return (1);
}

/* Check that the subject field contains only printing/space characters. */
static int
valid_subject(const char *subj)
{
	const char *p;
	
	for (p = subj; *p != '\0'; p++) {
		if (!isprint(*p))
			return (0);
	}
	return (1);
}

/* Feed the message to the mailer backend and archive it. */
static int
inject_mail(struct cgi_query *q, const char *sender, const char *rcpt,
    const char *subj, const char *message)
{
	char path[MAXPATHLEN];
	char stamp[FILENAME_MAX];
	const time_t clock = time(NULL);
	struct tm *tm = localtime(&clock);
	struct textbuf msg;
	struct stat sta;
	int fd, pdes[2], status;
	struct cgi_keyset *set;
	int i, eflag = 0;

	strlcpy(path, _PATH_MAIL_SENT, sizeof(path));
	if (stat(path, &sta) == -1 && mkdir(path, 0700) == -1) {
		error_set("%s: %s", path, strerror(errno));
		return (-1);
	}
	strftime(stamp, sizeof(stamp), "/%F.%T", tm);
	strlcat(path, stamp, sizeof(path));

	/* Construct the header of the message. */
	textbuf_init(&msg, MESSAGE_MAX, q->cset);
	textbuf_append_fmt(&msg, "From: %s\n", sender);
	textbuf_append_fmt(&msg, "To: %s\n", rcpt);
	textbuf_append_fmt(&msg, "Subject: %s\n", subj);
	textbuf_append_fmt(&msg, "X-MailFCGI-IP: %s\n", getenv("REMOTE_ADDR"));

	/* Append custom header fields. */
	set = cgi_get_keyset(q, "x_");
	for (i = 0; i < set->nents; i++) {
		char xkey[CGI_ARG_KEY_MAX];
		const char *xval;

		xkey[0] = 'x';
		xkey[1] = '_';
		strlcpy(&xkey[2], set->ents[i], sizeof(xkey)-2);

		xval = cgi_lget_arg(q, xkey, XVAL_MAX);
		textbuf_append_fmt(&msg, "X-%s: %s\n", set->ents[i], xval);
	}
	cgi_free_keyset(set);

	/* Append the message. */
	textbuf_append(&msg, message);
	textbuf_append(&msg, "\n");

	/* Append custom end of message fields. */
	set = cgi_get_keyset(q, "f_");
	for (i = 0; i < set->nents; i++) {
		char fkey[CGI_ARG_KEY_MAX];
		const char *fval;

		fkey[0] = 'f';
		fkey[1] = '_';
		strlcpy(&fkey[2], set->ents[i], sizeof(fkey)-2);
		
		if (!eflag) {
			eflag++;
			textbuf_append(&msg, "--\n");
		}

		fval = cgi_lget_arg(q, fkey, FVAL_MAX);
		textbuf_append_fmt(&msg, "%s: %s\n", set->ents[i], fval);
	}
	cgi_free_keyset(set);

	if ((fd = open(path, O_WRONLY|O_CREAT|O_EXCL, 0600)) == -1) {
		error_set("%s: %s", path, strerror(errno));
		goto fail;
	}
	if (write(fd, msg.buf, msg.len) < msg.len) {
		error_set("%s: write error", path);
		close(fd);
		goto fail;
	}
	close(fd);

	if (pipe(pdes) == -1) {
		error_set("pipe: %s", strerror(errno));
		goto fail;
	}
	if (fork()) {
		close(pdes[0]);
		if (write(pdes[1], msg.buf, msg.len) < msg.len ||
		    write(pdes[1], "\n", 1) < 1) {
			error_set("%s: write error", path);
			close(pdes[1]);
			goto fail;
		}
		close(pdes[1]);

		while (wait(&status) >= 0)
		    ;
		if (WIFEXITED(status) && WEXITSTATUS(status) != 0) {
			error_set("sendmail failure (%d)", WEXITSTATUS(status));
			goto fail;
		}
	} else {
		if (pdes[0] != STDIN_FILENO) {
			dup2(pdes[0], STDIN_FILENO);
			close(pdes[0]);
		}
		close(pdes[1]);
		execl(_PATH_SENDMAIL, "sendmail", "-i", rcpt, (char *)NULL);
		_exit(EX_UNAVAILABLE);
	}
	textbuf_destroy(&msg);
	return (0);
fail:
	textbuf_destroy(&msg);
	return (-1);
}

static int
mail_send(struct cgi_query *q)
{
	const char *sender, *rcptid, *subject, *message;
	const char *rcpt;

	if ((sender = cgi_lget_arg(q, "sender", SENDER_MAX)) == NULL ||
	    (rcptid = cgi_lget_arg(q, "rcpt", RCPT_MAX)) == NULL ||
	    (subject = cgi_lget_arg(q, "subject", SUBJECT_MAX)) == NULL ||
	    (message = cgi_lget_arg(q, "message", MESSAGE_MAX)) == NULL)
		goto fail;

	if (sender[0] == '\0' || !valid_sender(sender)) {
		error_set(_("The sender e-mail address is missing/invalid."));
		goto fail;
	}
	if ((rcpt = table_lookup_field(Trcpts, rcptid, 1)) == NULL) {
		error_set(_("The specified recipient does not exist."));
		goto fail;
	}
	if (subject[0] == '\0' || !valid_subject(subject)) {
		error_set(_("The subject field is missing/invalid."));
		goto fail;
	}
	if (message[0] == '\0') {
		error_set(_("The message is empty."));
		goto fail;
	}

	if (inject_mail(q, sender, rcpt, subject, message) == -1)
		goto fail;

	var_set("sender", "%s", sender);
	var_set("subject", "%s", subject);
	var_set("message", "%s", message);
	var_set("rcpt_name", "%s", table_lookup_field(Trcpts, rcptid, 2));
	html_output(q, "success");
	return (0);
fail:
	html_error(q, "%s", error_get());
	return (-1);
}

int
main(int argc, char *argv[])
{
	struct table_ent *ent;
	struct cgi_query *q;

	cgi_init(&mail_cgi);

	var_set_global("sender_max", "%d", SENDER_MAX);
	var_set_global("subject_max", "%d", SUBJECT_MAX);
	var_set_global("message_max", "%d", MESSAGE_MAX);
	var_set_global("xval_max", "%d", XVAL_MAX);
	var_set_global("fval_max", "%d", FVAL_MAX);

	if ((Trcpts = table_load("recipients", 3)) == NULL) {
		cgi_log(LOG_EMERG, "recipients: %s", error_get());
		return (1);
	}

	while ((q = cgi_accept()) != NULL) {
		const char *def_rcpt = cgi_lget_arg(q, "rcpt", RCPT_MAX);

		cgi_begin(q, "text/html");

		if (cgi_get_arg(q, "message") == NULL) {
			struct var *Vrcpts;
			
			Vrcpts = var_set("recipients", NULL);
			TABLE_FOREACH(ent, Trcpts) {
				const char *selected = "";

				if (strcmp(ent->fields[0], q->lang) != 0) {
					continue;
				}
				if (def_rcpt != NULL &&
				    strncmp(ent->key, def_rcpt,
				    strlen(ent->key)-3) == 0) {
					selected = " selected";
				}
				var_append(Vrcpts,
				    "<option value='%s'%s>%s</option>\n",
				    ent->key, selected, ent->fields[2]);
			}
			html_output(q, "default");
			goto out;
		}
		mail_send(q); 
out:
		cgi_end(q);
	}

	cgi_destroy(EX_OK, NULL);
	return (0);
}

static void
mail_destroy(void)
{
	table_destroy(Trcpts);
}

